import copy
import pickle
import pprint
from os import rename
from pathlib import Path
from time import time
from typing import Callable, Dict, Generator, Iterator, List, Optional, Union
from weakref import finalize

DataSet = Union[Iterator, Callable[..., Generator[dict, None, None]]]
WorkFlow = List[Dict[str, Union[Callable, str]]]
DataGenerator = Generator[dict, None, None]


class Data:
    """A wrapper for data/data_stream.

    The class provides methods for working with data as a stream.

    Such approach to data analysis called streaming transformation.
    """

    def __init__(self, data: DataSet, cache: bool = False, workflow: WorkFlow = None, parents_cache: List[str] = None):
        """
        Args:
            data: Data source.
            workflow: Workflow.
            parents_cache: Parents chain. Works as a stack.
            cache: Flag if you want write and read from cache.
        """
        self._cache_filename = f"{str(id(self))}:{time()}.pickle"
        self._len = None
        self._data = data
        self._workflow = [] if workflow is None else workflow
        self._length_hint = None  # The value is populated when we use limit method.
        self._limit_num = None
        self._cache_status = cache
        self._parents_cache = parents_cache if parents_cache else []
        self._finalizer = finalize(self, self.__remove)

    def __remove(self):
        """Data class destructor."""
        if self.__is_cache_file_exists(self._cache_filename):
            self.__delete_cache()
        del self._data

    def __delete_cache(self) -> None:
        """Removes cache file."""
        path = Path(self.__get_cache_filepath())
        path.unlink()

    @property
    def len(self) -> int:
        """int: How many records in the Data stream.

        Notes:
        1. It is a wasteful operation if you are performing it on the Data object that has never been iterated before.

        2. If you want just to check emptiness, use is_empty property instead.
        """
        return self._len if self._len is not None else self.__calc_len()

    @property
    def is_empty(self) -> bool:
        """bool: Indicates that the Data object doesn't contain data."""
        for _ in self.__load_data():
            return False
        return True

    def __calc_len(self) -> int:
        # TODO - request rpt-data-provide provide "select count"
        for _ in self:
            pass
        return self._len

    def __length_hint__(self):
        if self._len is not None:
            return self._len
        elif self._length_hint is not None:
            return self._length_hint
        else:
            # 2**13, though 8 - is a default value in CPython.
            # We usually have large number of data.
            return 8192

    def __iter__(self) -> DataSet:
        self._len = 0
        interruption = True
        try:
            for record in self.__load_data(self._cache_status):
                yield record
                self._len += 1
            else:
                # Loop fell through ..............
                interruption = False
        except StopIteration:
            return None
        finally:
            if interruption:
                if self.__is_cache_file_exists(self._cache_filename):
                    self.__delete_cache()

    def _build_workflow(self, workflow):
        new_workflow = copy.deepcopy(workflow)
        for w in new_workflow[::-1]:
            if w["type"] == "limit":
                w["callback"] = self._build_limit_callback(w["callback"].limit)

        return new_workflow

    def __load_data(self, cache: bool = False) -> DataGenerator:
        """Loads data from cache or data.

        Args:
            cache: Flag if you what write and read from cache.

        Returns:
            obj: Generator
        """
        if cache and self.__is_cache_file_exists(self._cache_filename):
            working_data = self.__load_file(self._cache_filename)
            yield from working_data
        else:
            cache_filename = self.get_last_cache()
            if cache_filename:
                working_data = self.__load_file(cache_filename)
                workflow = self.__get_unapplied_workflow(cache_filename)
            else:
                working_data = self._data() if callable(self._data) else self._data
                workflow = self._workflow

            workflow = self._build_workflow(workflow)

            if self.__check_file_recording():
                # Do not read from the cache file if it has PENDING status (if the file is not filled yet).
                cache = False

            yield from self.__change_data(working_data=working_data, workflow=workflow, cache=cache)

    def __check_file_recording(self) -> bool:
        """Checks whether there is a current recording in the file.

        Returns:
            bool: File recording status.
        """
        filename = f"[PENDING]{self._cache_filename}"
        return self.__is_cache_file_exists(filename)

    def __get_pending_cache_filepath(self) -> str:
        """Gets filepath for a cache file in pending status."""
        return f"./temp/[PENDING]{self._cache_filename}"

    def __get_cache_filepath(self) -> str:
        """Gets filepath for a cache file."""
        return f"./temp/{self._cache_filename}"

    def get_last_cache(self) -> Optional[str]:
        """Returns last existing cache.

        Returns: Cache filename
        """
        for cache_filename in self._parents_cache[::-1]:  # parents_cache works like a stack.
            if self.__is_cache_file_exists(cache_filename):
                return cache_filename
        return None

    def __get_unapplied_workflow(self, cache_filename) -> WorkFlow:
        """Returns list functions which haven't applied.

        Args:
            cache_filename: Cache filename in caches list of parents.

        Returns: Workflow which haven't applied.
        """
        cache_index = self._parents_cache[::-1].index(cache_filename)  # parents_cache works as a stack
        start_workflow = len(self._workflow) - 1 - cache_index  # each child has one more element then parent
        return self._workflow[start_workflow:]

    def __change_data(self, working_data: DataSet, workflow: WorkFlow, cache: bool = False) -> DataGenerator:
        """Applies workflow for data.

        Args:
            working_data: Data for apply workflow.
            workflow: Workflow.
            cache: Set True if you are going to write and read from the cache.

        Yields:
            obj: Generator
        """
        file = None
        if cache:
            filepath = self.__get_pending_cache_filepath()
            file = open(filepath, "wb")

        try:
            limit = False
            for record in working_data:
                modified_records = self.__apply_workflow(record, workflow)
                if modified_records is None:
                    break
                if not isinstance(modified_records, (list, tuple)):
                    modified_records = [modified_records]

                if None in modified_records:
                    limit = True
                    modified_records = modified_records[:-1]

                for modified_record in modified_records:
                    if file is not None:
                        pickle.dump(modified_record, file)
                    yield modified_record
                if limit:
                    break
        finally:
            if file:
                file.close()
                rename(file.name, self.__get_cache_filepath())

    def __is_cache_file_exists(self, filename: str) -> bool:
        """Checks whether file exist.

        Args:
            filename: Filename.

        Returns:
            File exists or not.

        """
        path = Path("./").joinpath("temp")
        path.mkdir(exist_ok=True)
        path = path.joinpath(filename)
        return path.is_file()

    def __load_file(self, filename: str) -> DataGenerator:
        """Loads records from pickle file.

        Args:
            filename: Filepath.

        Yields:
            obj: Generator records.

        """
        path = Path("./").joinpath("temp").joinpath(filename)
        if not path.exists():
            raise ValueError(f"{filename} doesn't exist.")

        if not path.is_file():
            raise ValueError(f"{filename} isn't file.")

        if path.suffix != ".pickle":
            raise ValueError(f"File hasn't pickle extension.")

        with open(path.resolve(), "rb") as file:
            while True:
                try:
                    decoded_data = pickle.load(file)
                    yield decoded_data
                except EOFError:
                    break

    def __apply_workflow(self, record: dict, workflow: WorkFlow) -> Optional[Union[dict, List[dict]]]:
        """Creates generator records with apply workflow.

        Returns:
            obj: Generator records.

        """
        for step in workflow:
            if isinstance(record, (list, tuple)):
                result = []
                for r in record:
                    try:
                        compute = step["callback"](r)
                        if compute is not None:
                            if not isinstance(compute, list):
                                compute = [compute]
                            result += compute
                    except StopIteration as e:
                        return [*result, None] if result else None

                record = result
                if not record:
                    record = None
                    break
            else:
                try:
                    record = step["callback"](record)
                except StopIteration as e:
                    return None
                if record is None:
                    break

        if record is None:
            record = []
        return record

    def filter(self, callback: Callable) -> "Data":
        """Append `filter` to workflow.

        Args:
            callback: Filter function.
                This function should return True or False.
                If function returns False, the record will be removed from the dataflow.

        Returns:
            Data: Data object.

        """
        new_workflow = [
            *self._workflow.copy(),
            {"type": "filter", "callback": lambda record: record if callback(record) else None},
        ]
        new_parents_cache = [*self._parents_cache, self._cache_filename]
        return Data(data=self._data, workflow=new_workflow, parents_cache=new_parents_cache)

    def map(self, callback: Callable) -> "Data":
        """Append `transform` function to workflow.

        Args:
            callback: Transform function.

        Returns:
            Data: Data object.

        """
        new_workflow = [*self._workflow.copy(), {"type": "map", "callback": callback}]
        new_parents_cache = [*self._parents_cache, self._cache_filename]
        return Data(data=self._data, workflow=new_workflow, parents_cache=new_parents_cache)

    def _build_limit_callback(self, num):
        def callback(r):
            if callback.pushed < num:
                callback.pushed += 1
                return r
            else:
                callback.pushed = 0
                raise StopIteration

        callback.limit = num
        callback.pushed = 0
        return callback

    def limit(self, num: int) -> "Data":
        """Limits the stream to `num` entries.

        Args:
            num: How many records will be provided.

        Returns:
            Data: Data object.

        """

        nwf = []
        for step in copy.deepcopy(self._workflow):
            if step["type"] == "limit":
                step["callback"] = self._build_limit_callback(step["callback"].limit)

            nwf.append(step)

        new_workflow = [*nwf, {"type": "limit", "callback": self._build_limit_callback(num)}]
        new_parents_cache = [*self._parents_cache, self._cache_filename]
        data_obj = Data(data=self._data, workflow=new_workflow, parents_cache=new_parents_cache)
        data_obj._length_hint = num
        data_obj._limit_num = num
        return data_obj

    def sift(self, limit: int = None, skip: int = None) -> Generator[dict, None, None]:
        """Skips and limits records.

        Args:
            limit: Limited records.
            skip: Skipped records.

        Yields:
            Generator records.

        """
        skipped = 0
        pushed = 0

        for record in self.__load_data():
            if skip is not None and skipped < skip:
                skipped += 1
                continue
            if limit is not None and pushed == limit:
                break
            yield record
            pushed += 1

    def use_cache(self, status: bool) -> "Data":
        """Change status cache.

        If True all requested data from rpt-data-provider will be saved to cache file.
        Further actions with Data object will be consume data from the cache file.

        Args:
            status(bool): Status.

        Returns:
            Data: Data object.

        """
        self._cache_status = status
        return self

    def find_by(self, record_field, field_values) -> Generator:
        """Get the records whose field value is written in the field_values list.

        When to use:
            You have IDs of some messages and you want get them in the stream and stop searching
            when you find all elements.

        Args:
            record_field: The record field to be searched for in the field_values list.
            field_values: List of elements among which will be searched record[record_field].

        Yields:
            dict: Generator records.

        """
        values_for_find = list(field_values)
        for record in self:
            if values_for_find:
                if record[record_field] in values_for_find:
                    values_for_find.remove(record[record_field])
                    yield record
                else:
                    continue
            else:
                break

    def write_to_file(self, file: str) -> None:
        """Writes the stream data to txt file.

        Args:
            file: Path to file.

        """
        with open(file, "w") as txt_file:
            for record in self:
                txt_file.write(f"{pprint.pformat(record)}\n" + ("-" * 50) + "\n")

    def __str__(self):
        output = "------------- Printed first 5 records -------------\n"
        for index, record in enumerate(self.__load_data()):
            if index == 5:
                break
            output += pprint.pformat(record) + "\n"
        return output

    def __bool__(self):
        for _ in self.__load_data():
            return True
        return False



from collections import defaultdict
from typing import Union, Iterator, Generator, Optional

from th2_data_services.data import Data
from th2_data_services.events_tree import EventsTree


class ParentEventsTree(EventsTree):
    """
    ParentEventsTree is a class (like an EventsTree).

    ParentEventsTree contains all parent events inside.
    """

    def __init__(self, data: Union[Iterator, Generator[dict, None, None], Data] = None, preserve_body: Optional[bool] = False):

        """
        Args:
            data: Events.
            preserve_body (:obj:`bool`, optional): if true keep events bodies.
        """
        if data is None:
            data = []

        self.__preserve_body = preserve_body
        self._parents_ids = set()
        self._unknown_events = defaultdict(lambda: 0)
        self._events = {}
        self.build_tree(data)

    @property
    def events(self) -> dict:
        return self._events

    def clear_events(self) -> None:
        """Clear exist events."""
        self._events.clear()

    @property
    def unknown_events(self):
        return self._unknown_events

    def clear_unknown_events(self) -> None:
        """Clear unknown events."""
        self._unknown_events.clear()

    def build_tree(self, data: Union[Iterator, Generator[dict, None, None]]) -> None:
        """Build parent events tree.

        :param data: Events.
        """
        for event in data:
            parent_id = event["parentEventId"]
            if parent_id is not None:
                self._parents_ids.add(parent_id)

        for event in data:
            self.append_element(event)
        self.search_unknown_parents()

    def append_element(self, event: dict) -> None:
        """Append new parent event to events tree.

        Args:
            event: Event
        """
        event_id = event["eventId"]
        if not self.__preserve_body:
            try:
                event.pop("body")
            except KeyError:
                pass

        if event_id in self._parents_ids:
            self._events[event_id] = event

        if event_id in self._unknown_events:
            self._unknown_events.pop(event_id)

    def search_unknown_parents(self) -> dict:
        """Searches unknown events.

        Returns:
             dict: Unknown events.
        """
        self.clear_unknown_events()
        for parent_id in self._parents_ids:
            if parent_id not in self.events.keys():
                self._unknown_events[parent_id] += 1
        return self._unknown_events



from codecs import register_error


def handler(err: UnicodeDecodeError):
    return chr(err.object[err.start]), err.end


UNICODE_REPLACE_HANDLER = "unicode_replace"

register_error(UNICODE_REPLACE_HANDLER, handler)



import requests
import json
import simplejson
import urllib3.exceptions
from itertools import chain
from requests.exceptions import ConnectionError
from urllib3 import PoolManager
from urllib.parse import quote
from functools import partial
from datetime import datetime, timezone
from csv import DictReader
from typing import Generator, Iterable, List, Union, Optional, Callable
from th2_data_services.sseclient import SSEClient

from th2_data_services.adapters import adapter_provider5, adapter_sse
from th2_data_services.data import Data
from http import HTTPStatus
from th2_data_services.filter import Filter
from th2_data_services.decode_error_handler import UNICODE_REPLACE_HANDLER

import logging

logger = logging.getLogger("th2_data_services")
logger.setLevel(logging.DEBUG)


class DataSource:
    """The class that provides methods for getting messages and events from rpt-data-provider."""

    def __init__(
        self,
        url: str,
        chunk_length: int = 65536,
        char_enc: str = "utf-8",
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
    ):
        """

        Args:
            url: HTTP data source url.
            chunk_length: How much of the content to read in one chunk.
            char_enc: Encoding for the byte stream.
            decode_error_handler: Registered decode error handler.
        """
        self.url = url
        self._char_enc = char_enc
        self._decode_error_handler = decode_error_handler
        self.__chunk_length = chunk_length
        self.__check_connect()
        logger.info(url)

    def __check_connect(self) -> None:
        """Checks whether url is working."""
        try:
            requests.get(self.__url, timeout=5.0)
        except ConnectionError as error:
            raise urllib3.exceptions.HTTPError(f"Unable to connect to host '{self.__url}'\nReason: {error}")

    @property
    def url(self) -> str:
        """str: URL of rpt-data-provider."""
        return self.__url

    @url.setter
    def url(self, url):
        if url[-1] == "/":
            url = url[:-1]
        self.__url = url

    def __get_data_obj(
        self,
        url: str,
        sse_adapter_flag: bool,
        provider_adapter: Optional[Callable],
        cache: bool,
    ) -> Data:
        data = partial(self.__execute_sse_request, url)

        if sse_adapter_flag:
            if provider_adapter is not None:
                data_obj = Data(data).map(adapter_sse).map(provider_adapter).use_cache(cache)
            else:
                data_obj = Data(data).map(adapter_sse).use_cache(cache)
        else:
            if provider_adapter is not None:
                raise Exception(f"Provider adapter expected to get dict but SSE adapter is turned off")
            else:
                data_obj = Data(data).use_cache(cache)

        return data_obj

    def sse_request_to_data_provider(self, **kwargs) -> Generator[dict, None, None]:
        """Sends SSE request to rpt-data-provider.

        It used for create custom sse-request to data-provider
        use this readme https://github.com/th2-net/th2-rpt-data-provider#readme.

        Args:
            kwargs: Query options.

        Yields:
            dict: SSE response data.

        """
        route = kwargs.get("route")
        if not route:
            exception_msg = "Route is required field. Please fill it."
            logger.error(exception_msg)
            raise ValueError(exception_msg)

        if kwargs.get("startTimestamp") and isinstance(kwargs.get("startTimestamp"), datetime):
            timestamp = kwargs["startTimestamp"].replace(tzinfo=timezone.utc).timestamp()
            kwargs["startTimestamp"] = int(timestamp * 1000)  # unix timestamp in milliseconds
        if kwargs.get("endTimestamp") and isinstance(kwargs.get("endTimestamp"), datetime):
            timestamp = kwargs["endTimestamp"].replace(tzinfo=timezone.utc).timestamp()
            kwargs["endTimestamp"] = int(timestamp * 1000)

        url = self.__url + route
        url = f"{url}?{self._get_url(kwargs)}"
        logger.info(url)
        yield from self.__execute_sse_request(url)

    def get_events_from_data_provider(self, cache: bool = False, sse_adapter: bool = True, **kwargs) -> Data:
        """Sends SSE request for getting events.

        For help use this readme
        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        on route http://localhost:8080/search/sse/events.

        Args:
            cache (bool): If True, all requested data from rpt-data-provider will be saved to cache.
            sse_adapter (bool): If True, all data will go through SSE adapter and yield dicts.
                Otherwise adapter will yield SSE Events.
            kwargs: th2-rpt-data-provider API query options.

        Returns:
            Data: Data object with Events.

        """

        if kwargs.get("metadataOnly") and kwargs["metadataOnly"] is True:
            logger.warning(f"Meta data status is {kwargs['metadataOnly']}")
        kwargs["metadataOnly"] = False

        if not kwargs.get("startTimestamp") and not kwargs.get("resumeFromId"):
            exception_msg = (
                "'startTimestamp' or 'resumeFromId' must not be null for route /search/sse/events. Please "
                "note it. More information on request here: "
                "https://github.com/th2-net/th2-rpt-data-provider "
            )
            logger.error(exception_msg)
            raise ValueError(exception_msg)

        if isinstance(kwargs["startTimestamp"], datetime):
            timestamp = kwargs["startTimestamp"].replace(tzinfo=timezone.utc).timestamp()
            kwargs["startTimestamp"] = int(timestamp * 1000)  # unix timestamp in milliseconds

        if kwargs.get("endTimestamp") and isinstance(kwargs.get("endTimestamp"), datetime):
            timestamp = kwargs["endTimestamp"].replace(tzinfo=timezone.utc).timestamp()
            kwargs["endTimestamp"] = int(timestamp * 1000)

        url = self.__url + "/search/sse/events"
        url = f"{url}?{self._get_url(kwargs)}"

        logger.info(url)

        return self.__get_data_obj(url, sse_adapter, None, cache)

    @staticmethod
    def _get_url(kwargs):
        result = ""
        for k, v in kwargs.items():
            if k == "filters":
                if isinstance(v, Filter):
                    result += v.url()
                elif isinstance(v, (tuple, list)):

                    result += "".join([filter_.url() for filter_ in v])
            else:
                result += f"&{k}={v}"
        url = result[1:] if result[0] == "&" else result
        return quote(url.encode(), "/:&?=")

    def get_messages_from_data_provider(
        self,
        cache: bool = False,
        sse_adapter: bool = True,
        provider_adapter: Optional[Callable] = adapter_provider5,
        **kwargs,
    ) -> Data:
        """Sends SSE request for getting messages.

        For help use this readme
        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        on route http://localhost:8080/search/sse/messages.

        Args:
            cache (bool): If True, all requested data from rpt-data-provider will be saved to cache.
            sse_adapter (bool): If True, all data will go through SSE adapter and yield dicts.
                Otherwise adapter will yield SSE Events.
            provider_adapter (callable): Adapter function for rpt-data-provider.
                If None, Data object will yield object from previous map function.
            kwargs: th2-rpt-data-provider API query options.

        Returns:
            Data: Data object with Messages.

        """
        if not kwargs.get("startTimestamp") and not kwargs.get("resumeFromId"):
            exception_msg = (
                "'startTimestamp' or 'resumeFromId' must not be null for route /search/sse/messages. "
                "Please note it. More information on request here: "
                "https://github.com/th2-net/th2-rpt-data-provider "
            )
            logger.error(exception_msg)
            raise ValueError(exception_msg)

        if not kwargs.get("stream"):
            exception_msg = (
                "'stream' is required field. Please note it."
                "More information on request here: "
                "https://github.com/th2-net/th2-rpt-data-provider "
            )
            logger.error(exception_msg)
            raise ValueError(exception_msg)

        if isinstance(kwargs["startTimestamp"], datetime):
            timestamp = kwargs["startTimestamp"].replace(tzinfo=timezone.utc).timestamp()
            kwargs["startTimestamp"] = int(timestamp * 1000)  # unix timestamp in milliseconds

        if kwargs.get("endTimestamp") and isinstance(kwargs.get("endTimestamp"), datetime):
            timestamp = kwargs["endTimestamp"].replace(tzinfo=timezone.utc).timestamp()
            kwargs["endTimestamp"] = int(timestamp * 1000)

        streams = kwargs.pop("stream")
        if not isinstance(streams, (list, tuple)):
            streams = [streams]

        url = self.__url + "/search/sse/messages"
        url = f"{url}?{self._get_url(kwargs)}"
        fixed_part_len = len(url)

        current_url, resulting_urls = "", []
        for stream in streams:
            stream = f"&stream={stream}"
            if fixed_part_len + len(current_url) + len(stream) >= 2048:
                resulting_urls.append(url + current_url)
                current_url = ""
            current_url += stream
        if current_url:
            resulting_urls.append(url + current_url)

        if len(resulting_urls) > 1:
            source = partial(
                chain.from_iterable,
                [self.__get_data_obj(res_url, sse_adapter, provider_adapter, cache) for res_url in resulting_urls],
            )
            return Data(source).use_cache(cache)
        else:
            url = resulting_urls[0]
            return self.__get_data_obj(url, sse_adapter, provider_adapter, cache)

    def __execute_sse_request(self, url: str) -> Generator[dict, None, None]:
        """Creates SSE connection to server.

        Args:
            url: Url.

        Returns:
            dict: Response data.

        """
        response = self.__create_stream_connection(url)
        client = SSEClient(response, char_enc=self._char_enc, decode_errors_handler=self._decode_error_handler)
        for record in client.events():
            yield record

    def __create_stream_connection(self, url: str) -> Generator[bytes, None, None]:
        """Create stream connection.

        Args:
            url: Url.

        Yields:
            str: Response stream data.

        """
        headers = {"Accept": "text/event-stream"}
        http = PoolManager()
        response = http.request(method="GET", url=url, headers=headers, preload_content=False)

        # Check response
        if response.status != HTTPStatus.OK:
            for s in HTTPStatus:
                if s == response.status:
                    raise urllib3.exceptions.HTTPError(f"{s.value} {s.phrase} ({s.description})")
            raise urllib3.exceptions.HTTPError(f"Http returned bad status: {response.status}")

        for chunk in response.stream(self.__chunk_length):
            yield chunk

        response.release_conn()

    def find_messages_by_id_from_data_provider(
        self,
        messages_id: Union[Iterable, str],
        provider_adapter: Optional[Callable] = adapter_provider5,
    ) -> Optional[Union[List[dict], dict]]:
        """Gets message/messages by ids.

        Args:
            messages_id: One str with MessageID or list of MessagesIDs.
            provider_adapter: Adapter function for rpt-data-provider.
                If None, the function return message/messages without applying the adapter function.

        Returns:
            List[Message_dict] if you request a list or Message_dict or None if no massages found.

        Example:
            >>> How to use.

            >>> data_source.find_messages_by_id_from_data_provider('demo-conn1:first:1619506157132265837')
            Returns 1 message (dict).

            >>> data_source.find_messages_by_id_from_data_provider(['demo-conn1:first:1619506157132265836'])
            Returns list(dict) with 1 message.

            >>> data_source.find_messages_by_id_from_data_provider([
                'demo-conn1:first:1619506157132265836',
                'demo-conn1:first:1619506157132265833',
            ])
            Returns list(dict) with 2 messages.

        """
        msg_id_type_is_str = isinstance(messages_id, str)
        if isinstance(messages_id, str):
            messages_id = [messages_id]

        result = []
        for msg_id in messages_id:
            index = None
            if msg_id.find(".") != -1:
                msg_id, index = "".join(msg_id.split(".")[:-1]), int(msg_id[-1])

            url = f"{self.__url}/message/{msg_id}"
            logger.info(url)
            response = requests.get(url)
            try:
                answer = response.json()
            except (json.JSONDecodeError, simplejson.JSONDecodeError):
                exception_msg = (
                    f"Sorry, but the answer rpt-data-provider doesn't match the json format.\n"
                    f"Answer:{response.text}"
                )
                logger.exception(exception_msg)
                raise ValueError(exception_msg)

            if provider_adapter is not None:
                answer = provider_adapter(answer)

            if isinstance(answer, list):
                if index:
                    for message in answer:
                        if message["body"]["metadata"]["id"]["subsequence"][0] == index:
                            result.append(message)
                            break
                else:
                    result += answer
            else:
                result.append(answer)

            if len(result) > 1:
                msg_id_type_is_str = False

        return result[0] if msg_id_type_is_str else result if result else None

    def find_events_by_id_from_data_provider(
        self, events_id: Union[Iterable, str], broken_events: Optional[bool] = False
    ) -> Optional[Union[List[dict], dict]]:
        """Gets event/events by ids.

        Args:
            events_id: One str with EventID or list of EventsIDs.
            broken_events: If True broken events is replaced by a event stub.

        Returns:
            List[Event_dict] if you request a list or Event_dict or None if no events found.

        Example:
            >>> How to use.

            >>> data_source.find_events_by_id_from_data_provider('8bc787fe-d1b4-11eb-bae5-57b0c4472880')
            Returns 1 message (dict).

            >>> data_source.find_events_by_id_from_data_provider(['8bc787fe-d1b4-11eb-bae5-57b0c4472880'])
            Returns list(dict) with 1 event.

            >>> data_source.find_events_by_id_from_data_provider([
                '8bc787fe-d1b4-11eb-bae5-57b0c4472880',
                '6e3be13f-cab7-4653-8cb9-6e74fd95ade4:8c035903-d1b4-11eb-9278-591e568ad66e',
            ])
            Returns list(dict) with 2 events.

        """

        def __create_event_stub(broken_event_id):
            """Creates a event stub.

            Args:
                broken_event_id: Broken event id.

            Returns:
                Event Stub.
            """
            event_stub = {
                "attachedMessageIds": [],
                "batchId": "Broken_Event",
                "endTimestamp": {"nano": 0, "epochSecond": 0},
                "startTimestamp": {"nano": 0, "epochSecond": 0},
                "type": "event",
                "eventId": f"{broken_event_id}",
                "eventName": "Broken_Event",
                "eventType": "Broken_Event",
                "parentEventId": "Broken_Event",
                "successful": None,
                "isBatched": None,
            }
            return event_stub

        def __get_event(event_id: str, stub: bool):
            """Gets event from rpt-data-provider or replace on event stub.

            Args:
                event_id: Event id.
                stub: If True a broken event is replaced by a event stub.
            """
            url = f"{self.__url}/event/{event_id}"
            logger.info(url)
            response = requests.get(url)

            try:
                return response.json()
            except (json.JSONDecodeError, simplejson.JSONDecodeError):
                if stub:
                    return __create_event_stub(event_id)
                exception_msg = (
                    f"Sorry, but the answer rpt-data-provider doesn't match the json format.\n"
                    f"Answer:{response.text}"
                )
                logger.error(exception_msg)
                raise ValueError(exception_msg)

        event_id_type_is_str = isinstance(events_id, str)
        if isinstance(events_id, str):
            events_id = [events_id]
        result = []
        for id_ in events_id:
            result.append(__get_event(id_, broken_events))

        return result[0] if event_id_type_is_str else result if result else None

    @staticmethod
    def read_csv_file(*sources: str) -> Generator[str, None, None]:
        """Gets data in a stream way from csv files.

        Args:
            sources: Path to files.

        Yields:
            dict: Csv files payload.

        """
        for source in sources:
            with open(source) as data:
                for message in DictReader(data):
                    yield dict(message)



from sseclient import Event
from urllib3.exceptions import HTTPError
import json


def adapter_sse(record: Event) -> dict:
    """SSE adapter.

    Args:
        record: SSE Event

    Returns:
        data dict
    """
    if record.event == "error":
        raise HTTPError(record.data)
    if record.event not in ["close", "keep_alive", "message_ids"]:
        try:
            return json.loads(record.data)
        except json.JSONDecodeError as e:
            raise Exception(f"json.decoder.JSONDecodeError: Invalid json received.\n" f"{e}\n" f"{record.data}")



from warnings import warn
from typing import Union, List


def adapter_provider5(record: dict) -> Union[List[dict], dict]:
    msg_type = record.get("messageType")
    if not msg_type:
        warn("Please note, some messages don't have a messageType field. Perhaps a codec doesn't decode some messages.", stacklevel=3)
        return record

    if "/" not in msg_type:
        return record

    body = record["body"]
    if not body:
        return record

    sub_messages = []
    fields = body["fields"]
    if not fields:
        return record

    for sub_msg in fields:
        split_msg_name = sub_msg.split("-")
        if len(split_msg_name) > 1:
            sub_msg_type, index = "".join(split_msg_name[:-1]), int(split_msg_name[-1])
        else:
            index = msg_type.split("/").index(sub_msg) + 1
            sub_msg_type = sub_msg

        new_record = record.copy()

        metadata = new_record["body"]["metadata"].copy()
        id_field = metadata["id"].copy()
        id_field["subsequence"] = [index]
        metadata["id"] = id_field

        body_fields = fields[sub_msg]
        metadata.update(body_fields.get("metadata", {}))

        body = {"metadata": metadata}
        if body_fields.get("messageValue"):
            body = {**body_fields["messageValue"], **body}
        elif body_fields.get("fields"):
            body = {**body_fields["fields"], **body}
        else:
            body = {"fields": {}, **body}

        new_record["body"] = body
        new_record["body"]["metadata"]["messageType"] = sub_msg_type
        new_record["messageType"] = sub_msg_type
        new_record["messageId"] = f"{record['messageId']}.{index}"
        sub_messages.append(new_record)
    return sub_messages
